/*
 * Copyright 2012 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.gradle.api.internal.project.taskfactory

import org.gradle.api.DefaultTask
import org.gradle.api.InvalidUserDataException
import org.gradle.api.Task
import org.gradle.api.internal.AbstractTask
import org.gradle.api.internal.ClassGenerator
import org.gradle.api.internal.TaskInternal
import org.gradle.api.tasks.TaskInstantiationException
import org.gradle.internal.reflect.Instantiator
import org.gradle.api.reflect.ObjectInstantiationException
import org.gradle.test.fixtures.AbstractProjectBuilderSpec

class TaskFactoryTest extends AbstractProjectBuilderSpec {
    final ClassGenerator generator = Mock()
    final Instantiator instantiator = Mock()
    ITaskFactory taskFactory

    def setup() {
        taskFactory = new TaskFactory(generator).createChild(project, instantiator)
        _ * generator.generate(_) >> { Class type -> type }
        _ * instantiator.newInstance(_) >> { args -> args[0].newInstance() }
    }

    public void injectsProjectAndNameIntoTask() {
        when:
        Task task = taskFactory.create("task", DefaultTask)

        then:
        task.project == project
        task.name == 'task'
    }

    public void testCreateTaskOfTypeWithNoArgsConstructor() {
        when:
        Task task = taskFactory.create('task', TestDefaultTask.class)

        then:
        task instanceof TestDefaultTask
    }

    public void testCreateTaskWhereSuperTypeOfDefaultImplementationRequested() {
        when:
        Task task = taskFactory.create('task', type)

        then:
        task instanceof DefaultTask

        where:
        type << [Task, TaskInternal, AbstractTask, DefaultTask]
    }

    public void instantiatesAnInstanceOfTheDecoratedTaskType() {
        when:
        Task task = taskFactory.create('task', TestDefaultTask.class)

        then:
        task instanceof DecoratedTask

        and:
        1 * generator.generate(TestDefaultTask) >> DecoratedTask
        1 * instantiator.newInstance(DecoratedTask) >> { new DecoratedTask() }
        0 * _._
    }

    public void testCreateTaskForTypeWhichDoesNotImplementTask() {
        when:
        taskFactory.create('task', NotATask)

        then:
        InvalidUserDataException e = thrown()
        e.message == "Cannot create task of type 'NotATask' as it does not implement the Task interface."
    }

    public void wrapsFailureToCreateTaskInstance() {
        def failure = new RuntimeException()

        when:
        taskFactory.create('task', TestDefaultTask)

        then:
        TaskInstantiationException e = thrown()
        e.message == "Could not create task of type 'TestDefaultTask'."
        e.cause == failure

        and:
        _ * instantiator.newInstance(TestDefaultTask) >> { throw new ObjectInstantiationException(TestDefaultTask, failure) }
    }

    public static class TestDefaultTask extends DefaultTask {
    }

    public static class DecoratedTask extends TestDefaultTask {
    }

    public static class NotATask {
    }
}
